{ Unit: BcDialogs
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Enchanced CommonDialog components to allow easy adding of custom
    controls to the dialogs through the OnAddControls event.

    Introduces properties to change existing controls.

  History:

    2001-10-03: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcDialogs;

interface

uses
  { For Delphi 6 compatibility Forms must come after Classes }
  Windows, Messages, SysUtils, Classes, Controls, Dialogs,
  CommDlg, Contnrs, Dlgs, StdCtrls, Forms;

type
  TDialogControl = (dcOkButton, dcCancelButton, dcHelpButton, dcFileNameLabel,
    dcFileNameEdit, dcFileTypeLabel, dcFileTypeCombobox, dcFolderLabel, dcFolderCombobox,
    dcButtonToolbar, {dcFileView,} dcPlacesToolbar);
  TDialogControls = set of TDialogControl;

  TShareAction = (saDefault, saAccept, saRejectNoWarning);
  TShareViolationEvent = procedure(const AFileName: string;
    var ShareAction: TShareAction) of object;

const
  DLGID_OK              = $0001;
  DLGID_CANCEL          = $0002;
  DLGID_HELP            = pshHelp;
  DLGID_FNLABEL         = stc3;
  DLGID_FNEDIT          = edt1;
  DLGID_FTLABEL         = stc2;
  DLGID_FTCOMBOBOX      = cmb1;
  DLGID_FOLDERLABEL     = stc4;
  DLGID_FOLDERCOMBOBOX  = cmb2;

  DLGID_FILEVIEW        = lst2;
  DLGID_BUTTONTOOLBAR   = 1;
  DLGID_PLACESTOOLBAR   = $04A0;

  DialogControlToDLGID: array[TDialogControl] of Longint =
    (DLGID_OK, DLGID_CANCEL, DLGID_HELP, DLGID_FNLABEL, DLGID_FNEDIT,
     DLGID_FTLABEL, DLGID_FTCOMBOBOX, DLGID_FOLDERLABEL, DLGID_FOLDERCOMBOBOX,
     DLGID_BUTTONTOOLBAR, {DLGID_FILEVIEW,} DLGID_PLACESTOOLBAR);

type
  TBcOpenDialog = class;

  TBcDialogCaptions = class(TPersistent)
  private
    FHelpButton: string;
    FCancelButton: string;
    FFileTypeLabel: string;
    FFileNameLabel: string;
    FOkButton: string;
    FFolderLabel: string;
    FOwner: TBcOpenDialog;
    procedure SetCancelButton(const Value: string);
    procedure SetFileNameLabel(const Value: string);
    procedure SetFolderLabel(const Value: string);
    procedure SetHelpButton(const Value: string);
    procedure SetOkButton(const Value: string);
    procedure SetFileTypeLabel(const Value: string);
  public
    constructor Create(AOwner: TBcOpenDialog);
    property Owner: TBcOpenDialog read FOwner;
  published
    property OkButton: string read FOkButton write SetOkButton;
    property CancelButton: string read FCancelButton write SetCancelButton;
    property HelpButton: string read FHelpButton write SetHelpButton;
    property FileNameLabel: string read FFileNameLabel write SetFileNameLabel;
    property FileTypeLabel: string read FFileTypeLabel write SetFileTypeLabel;
    property FolderLabel: string read FFolderLabel write SetFolderLabel;
  end;

  TBcOpenDialog = class(TOpenDialog)
  private
    FOnCreate: TNotifyEvent;
    FOnDestroy: TNotifyEvent;
    FCreatedControls: TObjectList;
    FOnAddControls: TNotifyEvent;
    FVisibleControls: TDialogControls;
    FCaptions: TBcDialogCaptions;
    FOnHelp: THelpEvent;
    FOnHelpClick: TNotifyEvent;
    FOnShareViolation: TShareViolationEvent;
    function GetClientRect: TRect;
    function GetCreatedControls: TObjectList;
    function GetWinPos(const Index: Integer): Integer;
    procedure SetWinPos(const Index, Value: Integer);
    function GetWindowRect: TRect;
    procedure SetCaptions(const Value: TBcDialogCaptions);
    function GetParentHandle: THandle;
    procedure SetWindowRect(const Value: TRect);
    function GetDialogControlRect(Item: TDialogControl): TRect;
    function GetFileName: TFileName;
    procedure SetFileName(const Value: TFileName);
    { Private declarations }
  protected
    { Protected declarations }
    function GetStaticRect: TRect; override;
    procedure DoCreate; dynamic;
    procedure DoDestroy; dynamic;
    function DoAddControls: Boolean; dynamic;
    procedure DoShow; override;

    procedure WndProc(var Message: TMessage); override;

    procedure SetControlText(DialogControl: TDialogControl; const Caption: string);
    procedure HideControl(DialogControl: TDialogControl);
    procedure HideClassControl(AClassName: string; DialogControl: TDialogControl; ParentHandle: THandle = 0);

    property CreatedControls: TObjectList read GetCreatedControls;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure AddControl(Control: TControl; const X, Y: Integer);
    function AddButton(const ACaption: string; const AOnClick: TNotifyEvent;
      const X, Y: Integer; const AWidth: Integer; const AHeight: Integer): TButton;
    function AddLabel(const ACaption: string; const X, Y: Integer): TLabel;

    procedure ClickOk;
    procedure ClickCancel;

    procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); virtual;

    function Execute: Boolean; override;

    property StaticRect: TRect read GetStaticRect;
    property ClientRect: TRect read GetClientRect;
    property WindowRect: TRect read GetWindowRect write SetWindowRect;

    property Left: Integer index 0 read GetWinPos write SetWinPos;
    property Top: Integer index 1 read GetWinPos write SetWinPos;
    property Width: Integer index 2 read GetWinPos write SetWinPos;
    property Height: Integer index 3 read GetWinPos write SetWinPos;

    property ParentHandle: THandle read GetParentHandle;

    property DialogControlRect[Item: TDialogControl]: TRect read GetDialogControlRect;
  published
    { Published declarations }

    { Properties }
    property FileName: TFileName read GetFileName write SetFileName;
    property VisibleControls: TDialogControls read FVisibleControls write FVisibleControls;
    property Captions: TBcDialogCaptions read FCaptions write SetCaptions;

    { Events }
    property OnCreate: TNotifyEvent read FOnCreate write FOnCreate;
    property OnDestroy: TNotifyEvent read FOnDestroy write FOnDestroy;
    property OnAddControls: TNotifyEvent read FOnAddControls write FOnAddControls;
    property OnHelp: THelpEvent read FOnHelp write FOnHelp;
    property OnHelpClick: TNotifyEvent read FOnHelpClick write FOnHelpClick;
    property OnShareViolation: TShareViolationEvent read FOnShareViolation write FOnShareViolation;
  end;

  TBcSaveDialog = class(TBcOpenDialog)
  public
    function Execute: Boolean; override;
  end;

implementation

uses Buttons, extctrls;

{ TBcSaveDialog }

function TBcSaveDialog.Execute: Boolean;
begin
  Result := DoExecute(@GetSaveFileName);
end;

{ TBcOpenDialog }

function TBcOpenDialog.AddButton(const ACaption: string; const AOnClick: TNotifyEvent;
  const X, Y: Integer; const AWidth: Integer; const AHeight: Integer): TButton;
var
  Btn: TButton;
{  Pnl: TPanel;}
begin
  Result := nil;
  if Handle <> 0 then
  begin

  {  Pnl := TPanel.Create(Self);
    Pnl.Caption := '';
    Pnl.SetBounds(X, Y, AWidth, AHeight);
    Pnl.BorderWidth := 0;
    Pnl.BevelOuter := bvNone;}

    Btn := TButton.Create({Pnl}Self);
    Btn.Caption := ACaption;
    Btn.OnClick := AOnClick;
    Btn.SetBounds(0, 0, AWidth, AHeight);

  {  Btn.Parent := Pnl;}

  {  CreatedControls.Add(Pnl);
    CreatedControls.Add(Btn);
    Pnl.ParentWindow := GetParent(Handle);
    SetWindowPos(Pnl.Handle, 0, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);}

    AddControl(Btn, X, Y);

    Result := Btn;
  end;
end;

procedure TBcOpenDialog.AddControl(Control: TControl; const X, Y: Integer);
var
  Pnl: TPanel;
begin
  if Handle <> 0 then
  begin
    Pnl := TPanel.Create(Self);
    Pnl.Caption := '';
    Pnl.SetBounds(X, Y, Control.Width, Control.Height);
    Pnl.BorderWidth := 0;
    Pnl.BevelOuter := bvNone;

    Control.SetBounds(0, 0, Control.Width, Control.Height);
    Control.Parent := Pnl;

    CreatedControls.Add(Pnl);
    CreatedControls.Add(Control);
    Pnl.ParentWindow := GetParent(Handle);
    SetWindowPos(Pnl.Handle, 0, 0, 0, 0, 0, SWP_NOMOVE or SWP_NOSIZE);
  end;
end;

function TBcOpenDialog.Execute: Boolean;
begin
  Result := inherited Execute;

  FreeAndNil(FCreatedControls);
end;

procedure TBcOpenDialog.WndProc(var Message: TMessage);
const
  ShareResult: array[TShareAction] of Cardinal =
    (0, OFN_SHAREFALLTHROUGH, OFN_SHARENOWARN);
var
  CallHelp: Boolean;
  ShareAction: TShareAction;
  AFileName: string;
begin
  CallHelp := False;
  if (Message.Msg = WM_NOTIFY) then
    case (POFNOTIFY(Message.LParam)^.hdr.code) of
      CDN_HELP:
        begin
          if Assigned(FOnHelp) and (HelpContext <> 0) then
            FOnHelp(HELP_CONTEXT, HelpContext, CallHelp);
          if Assigned(FOnHelpClick) then
            FOnHelpClick(Self);
        end;
      CDN_SHAREVIOLATION:
        begin
          if Assigned(FOnShareViolation) then
          begin
            ShareAction := saDefault;
            AFileName := PAnsiChar((POFNOTIFY(Message.LParam)^.pszFile));
            FOnShareViolation(AFileName, ShareAction);
            Message.Result := ShareResult[ShareAction];
            if (ShareAction <> saDefault) then
              SetWindowLong(Handle, DWL_MSGRESULT, Message.Result);
          end;
        end;
    end;
  inherited WndProc(Message);
end;

function TBcOpenDialog.GetParentHandle: THandle;
begin
  if Handle <> 0 then
    Result := GetParent(Handle)
  else
    Result := 0;
end;

procedure TBcOpenDialog.SetWindowRect(const Value: TRect);
begin
  if Handle <> 0 then
    SetWindowPos(GetParent(Handle), 0, Value.Left, Value.Top,
      Value.Right - Value.Left, Value.Bottom - Value.Top,
        SWP_NOACTIVATE or SWP_NOOWNERZORDER	or SWP_NOZORDER);
end;

function TBcOpenDialog.AddLabel(const ACaption: string; const X,
  Y: Integer): TLabel;
var
  Lbl: TLabel;
begin
  Result := nil;
  if Handle <> 0 then
  begin
    Lbl := TLabel.Create(Self);
    Lbl.Caption := ACaption;
    AddControl(Lbl, X, Y);
    Result := Lbl;
  end;
end;

function TBcOpenDialog.GetDialogControlRect(Item: TDialogControl): TRect;
begin
  Windows.GetWindowRect(GetDlgItem(GetParent(Handle), DialogControlToDLGID[Item]), Result);
  MapWindowPoints(0, GetParent(Handle), Result, 2);
end;

function TBcOpenDialog.GetFileName: TFileName;
begin
  Result := inherited FileName;
end;

procedure TBcOpenDialog.SetFileName(const Value: TFileName);
begin
  inherited FileName := Value;
  if Handle <> 0 then
    SetControlText(dcFileNameEdit, ExtractFileName(Value));
end;

{ TBcDialogCaptions }

constructor TBcDialogCaptions.Create(AOwner: TBcOpenDialog);
begin
  inherited Create;
  FOwner := AOwner;
end;

procedure TBcDialogCaptions.SetCancelButton(const Value: string);
begin
  FCancelButton := Value;
  if Owner.Handle <> 0 then
    Owner.SetControlText(dcCancelButton, Value);
end;

procedure TBcDialogCaptions.SetFileNameLabel(const Value: string);
begin
  FFileNameLabel := Value;
  if Owner.Handle <> 0 then
    Owner.SetControlText(dcFileNameLabel, Value);
end;

procedure TBcDialogCaptions.SetFileTypeLabel(const Value: string);
begin
  FFileTypeLabel := Value;
  if Owner.Handle <> 0 then
    Owner.SetControlText(dcFileTypeLabel, Value);
end;

procedure TBcDialogCaptions.SetFolderLabel(const Value: string);
begin
  FFolderLabel := Value;
  if Owner.Handle <> 0 then
    Owner.SetControlText(dcFolderLabel, Value);
end;

procedure TBcDialogCaptions.SetHelpButton(const Value: string);
begin
  FHelpButton := Value;
  if Owner.Handle <> 0 then
    Owner.SetControlText(dcHelpButton, Value);
end;

procedure TBcDialogCaptions.SetOkButton(const Value: string);
begin
  FOkButton := Value;
  if Owner.Handle <> 0 then
    Owner.SetControlText(dcOkButton, Value);
end;

procedure TBcOpenDialog.ClickCancel;
begin
  if Handle <> 0 then
    SendMessage(GetParent(Handle), WM_CLOSE, 0, 0);
end;

procedure TBcOpenDialog.ClickOk;
begin
  if Handle <> 0 then
  begin
    SetFocus(GetDlgItem(GetParent(Handle), DLGID_OK));
    keybd_event(VK_RETURN, 0, 0, 0);
    keybd_event(VK_RETURN, 0, KEYEVENTF_KEYUP, 0);
  end;
end;

constructor TBcOpenDialog.Create(AOwner: TComponent);
var I: TDialogControl;
begin
  inherited Create(AOwner);
  for I := Low(TDialogControl) to High(TDialogControl) do
    Include(FVisibleControls, I);

  FCaptions := TBcDialogCaptions.Create(Self);
  DoCreate;
end;

destructor TBcOpenDialog.Destroy;
begin
  DoDestroy;
  FreeAndNil(FCaptions);
  FreeAndNil(FCreatedControls);
  inherited Destroy;
end;

function TBcOpenDialog.DoAddControls: Boolean;
begin
  Result := True;
  if Assigned(FOnAddControls) then
    FOnAddControls(Self);
end;

procedure TBcOpenDialog.DoCreate;
begin
  if Assigned(FOnCreate) then
    FOnCreate(Self);
end;

procedure TBcOpenDialog.DoDestroy;
begin
  if Assigned(FOnDestroy) then
    FOnDestroy(Self);
end;

procedure TBcOpenDialog.DoShow;
begin
  if DoAddControls then
  begin
    if dcOkButton in FVisibleControls then
      SetControlText(dcOkButton, Captions.OkButton)
    else
      HideControl(dcOkButton);

    if dcCancelButton in FVisibleControls then
      SetControlText(dcCancelButton, Captions.CancelButton)
    else
      HideControl(dcCancelButton);

    if dcHelpButton in FVisibleControls then
      SetControlText(dcHelpButton, Captions.HelpButton)
    else
      HideControl(dcHelpButton);

    if dcFileNameLabel in FVisibleControls then
      SetControlText(dcFileNameLabel, Captions.FileNameLabel)
    else
      HideControl(dcFileNameLabel);

    if dcFileTypeLabel in FVisibleControls then
      SetControlText(dcFileTypeLabel, Captions.FileTypeLabel)
    else
      HideControl(dcFileTypeLabel);

    if dcFolderLabel in FVisibleControls then
      SetControlText(dcFolderLabel, Captions.FolderLabel)
    else
      HideControl(dcFolderLabel);

    if not (dcFolderCombobox in FVisibleControls) then
      HideControl(dcFolderCombobox);

    if not (dcFileNameEdit in FVisibleControls) then
      HideControl(dcFileNameEdit);

    if not (dcFileTypeCombobox in FVisibleControls) then
      HideControl(dcFileTypeCombobox);

//    if not (dcFileView in FVisibleControls) then
//      HideClassControl('SHELLDLL_DefView', dcFileView);

    if not (dcButtonToolbar in FVisibleControls) then
      HideClassControl('ToolbarWindow32', dcButtonToolbar);

    if not (dcPlacesToolbar in FVisibleControls) then
      HideClassControl('ToolbarWindow32', dcPlacesToolbar);
  end;
  inherited DoShow;
end;

function TBcOpenDialog.GetClientRect: TRect;
begin
  Windows.GetClientRect(GetParent(Handle), Result);
end;

function TBcOpenDialog.GetCreatedControls: TObjectList;
begin
  if not Assigned(FCreatedControls) then
  begin
    FCreatedControls := TObjectList.Create;
    FCreatedControls.OwnsObjects := True;
  end;
  Result := FCreatedControls;
end;

function TBcOpenDialog.GetStaticRect: TRect;
begin
  Result := inherited GetStaticRect;
end;

function TBcOpenDialog.GetWindowRect: TRect;
begin
  if Handle <> 0 then
    Windows.GetWindowRect(GetParent(Handle), Result)
  else
    Result := Rect(0, 0, 0, 0);
end;

function TBcOpenDialog.GetWinPos(const Index: Integer): Integer;
var R: TRect;
begin
  R := WindowRect;
  case Index of
    0: Result := R.Left;
    1: Result := R.Top;
    2: Result := R.Right - R.Left;
    3: Result := R.Bottom - R.Top;
  else
    Result := 0;
  end;
end;

procedure TBcOpenDialog.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
  SetWindowPos(GetParent(Handle), 0, ALeft, ATop, AWidth, ATop,
    SWP_NOACTIVATE or SWP_NOOWNERZORDER);
end;

procedure TBcOpenDialog.SetCaptions(const Value: TBcDialogCaptions);
begin
  FCaptions.Assign(Value);
end;

procedure TBcOpenDialog.SetWinPos(const Index, Value: Integer);
var R: TRect;
begin
  R := WindowRect;
  case Index of
    0: R.Left := Value;
    1: R.Top := Value;
    2: R.Right := {R.Left + }Value;
    3: R.Bottom := {R.Top + }Value;
  end;
  SetBounds(R.Left, R.Top, R.Right, R.Bottom);
end;

procedure TBcOpenDialog.SetControlText(DialogControl: TDialogControl; const Caption: string);
begin
  if Caption <> '' then
    SendMessage(GetParent(Handle), CDM_SETCONTROLTEXT,
      DialogControlToDLGID[DialogControl], Longint(PChar(Caption)));
end;

procedure TBcOpenDialog.HideControl(DialogControl: TDialogControl);
begin
  SendMessage(GetParent(Handle), CDM_HIDECONTROL,
    DialogControlToDLGID[DialogControl], 0);
end;

procedure TBcOpenDialog.HideClassControl(AClassName: string; DialogControl: TDialogControl; ParentHandle: THandle = 0);
var
  FindHandle: THandle;
begin
  if ParentHandle = 0 then
    ParentHandle := GetParent(Handle);
  FindHandle := FindWindowEx(ParentHandle, 0, PChar(AClassName), nil);
  while (FindHandle <> 0) and (GetDlgCtrlID(FindHandle) <> DialogControlToDLGID[DialogControl]) do
    FindHandle := FindWindowEx(ParentHandle, FindHandle, PChar(AClassName), nil);
  if (FindHandle <> 0) then
    ShowWindow(FindHandle, SW_HIDE);
end;

initialization

finalization

end.
